from cerberus import Validator
from cerberus.errors import ValidationError
import json, logging

messenger_backend_settings_schema = {
    'RABBITMQ_MANAGEMENT_SERVER': {'type': 'string', 'required': True,},
    'RABBITMQ_ADMIN_USER': {'type': 'string', 'required': True,},
    'RABBITMQ_ADMIN_PASSWORD': {'type': 'string', 'required': True,},
    'MQTT_BROKER': {'type': 'string', 'required': True,},
}

orsim_settings_schema = {
    'SIMULATION_LENGTH_IN_STEPS': {'type': 'integer', 'required': True,},
    'STEP_INTERVAL': {'type': 'integer', 'required': True,},

    'AGENT_LAUNCH_TIMEOUT': {'type': 'integer', 'required': True,},
    'STEP_TIMEOUT': {'type': 'integer', 'required': True,},
    'STEP_TIMEOUT_TOLERANCE': {'type': 'float', 'required': True,}, # NOTE deprecated

    'REFERENCE_TIME': {'type': 'string', 'required': True,},
}

class ORSimEnv:

    messenger_settings = None

    @classmethod
    def set_backend(cls, settings):
        v = Validator(allow_unknown=True)

        if v.validate(settings, messenger_backend_settings_schema):
            cls.messenger_settings = settings
        else:
            logging.error(f'{json.dumps(v.errors, indent=2)}')
            raise ValidationError(json.dumps(v.errors))


    @classmethod
    def validate_orsim_settings(cls, settings):
        v = Validator(allow_unknown=True)

        if v.validate(settings, orsim_settings_schema):
            return settings
        else:
            logging.error(f'{json.dumps(v.errors, indent=2)}')
            raise ValidationError(json.dumps(v.errors))



import string
import random
from  collections.abc import Mapping

from datetime import datetime


# def id_generator(size=12, chars=string.ascii_lowercase + string.ascii_uppercase + string.digits):
#     return ''.join(random.SystemRandom().choice(chars) for _ in range(size))


# def is_success(status_code):
#     return (status_code >= 200) and (status_code <= 299)


# def deep_update(source, overrides):
#     """
#     Update a nested dictionary or similar mapping.
#     Modify ``source`` in place.
#     """
#     for key, value in overrides.items():
#         if isinstance(value, Mapping) and value:
#             returned = deep_update(source.get(key, {}), value)
#             source[key] = returned
#         else:
#             source[key] = overrides[key]
#     return source


def time_to_str(time_var):
    return datetime.strftime(time_var, "%a, %d %b %Y %H:%M:%S GMT")

def str_to_time(time_str):
    return datetime.strptime(time_str, "%a, %d %b %Y %H:%M:%S GMT")




from abc import ABC, abstractclassmethod, abstractmethod
import asyncio, json, logging, time, os, pprint

from datetime import datetime
# from apps import orsim
from orsim.messenger import Messenger

from random import random
from .orsim_env import ORSimEnv

from orsim.tasks import start_agent


class ORSimScheduler(ABC):

    def __init__(self, run_id, scheduler_id, orsim_settings, init_failure_handler='soft'):

        if ORSimEnv.messenger_settings is None:
            raise Exception("Please Initialize ORSimEnv.set_backend()")

        self.orsim_settings = ORSimEnv.validate_orsim_settings(orsim_settings)

        self.run_id = run_id
        self.scheduler_id = scheduler_id
        self.time = 0

        self.init_failure_handler = init_failure_handler

        self.agent_collection = {}
        self.agent_stat = {}

        self.agent_credentials = {
            'email': f"{self.run_id}_{self.scheduler_id}_ORSimScheduler",
            'password': "secret_password",
        }

        self.agent_messenger = Messenger(ORSimEnv.messenger_settings, self.agent_credentials, f"{self.run_id}/{self.scheduler_id}/ORSimScheduler", self.on_receive_message)

        self.pp = pprint.PrettyPrinter(indent=2)
        logging.info(f'Starting new {scheduler_id= } for {run_id= }')


    # def add_agent(self, unique_id, method, spec):
    def add_agent(self, spec, #unique_id,
                  project_path, agent_class):
        ''' '''
        self.agent_collection[spec['unique_id']] = {
            # 'method': method,
            'spec': spec,
            # 'step_response': 'waiting'
            'step_response': {
                self.time: {
                    'reaction': 'waiting',
                    'did_step': False,
                    'run_time': 0,
                }
            }
        }

        # import sys
        # if not project_path in sys.path:
        #     sys.path.append(project_path)
        # print(sys.path)

        module_comp = agent_class.split('.')
        module_name, agent_class_name = str.join('.', module_comp[:-1]), module_comp[-1]

        kwargs = spec.copy()
        # kwargs['scheduler_id'] = self.scheduler_id
        kwargs['scheduler'] = {
            'id': self.scheduler_id,
            'orsim_settings': self.orsim_settings
        }
        # method.delay(**kwargs) # NOTE This starts the Celery Task in a new worker thread
        start_agent.delay(project_path, module_name, agent_class_name, ORSimEnv.messenger_settings, **kwargs) # NOTE This starts the Celery Task in a new worker thread

        logging.info(f"agent {spec['unique_id']} entering market")
        print(f"agent {spec['unique_id']} entering market")

        # launch_start = time.time()
        # while True:
        #     # if self.agent_collection[unique_id]['step_response'] == 'ready':
        #     if self.agent_collection[unique_id]['step_response'][self.time]['reaction'] == 'ready':
        #         logging.info(f'agent {unique_id} is ready')
        #         print(f'agent {unique_id} is ready')
        #         break
        #     elif (self.agent_collection[unique_id]['step_response'][self.time]['reaction'] == 'init_error') or \
        #                         ((time.time() - launch_start) > self.orsim_settings['AGENT_LAUNCH_TIMEOUT']):
        #         logging.exception(f'Failed to Launch agent {unique_id}')
        #         print(f'Failed to Launch agent {unique_id}')
        #         self.remove_agent(unique_id)
        #         if self.init_failure_handler == 'soft':
        #             break
        #         else:
        #             raise Exception(f"Shutdown {self.scheduler_id} due to {self.init_failure_handler=}. Agent {unique_id} failed to launch.")
        #     else:
        #         time.sleep(0.1)


    def remove_agent(self, unique_id):
        try:
            logging.info(f"agent {unique_id} has left")
            print(f"agent {unique_id} has left")
            self.agent_collection.pop(unique_id)
        except Exception as e:
            logging.exception(str(e))
            # print(e)

    def on_receive_message(self, client, userdata, message):
        if message.topic == f"{self.run_id}/{self.scheduler_id}/ORSimScheduler":
            payload = json.loads(message.payload.decode('utf-8'))

            response_time_step = payload.get('time_step') if payload.get('time_step') != -1 else self.time

            try:
                self.agent_collection[payload.get('agent_id')]['step_response'][response_time_step] = {
                    'reaction': payload.get('action'),
                    'did_step': payload.get('did_step'),
                    'run_time': payload.get('run_time')
                }
            except: pass
            # except Exception as e:
            #     logging.exception(str(e))

            if (payload.get('action') == 'error') or (response_time_step is None):
                logging.warning(f'{self.__class__.__name__} received {message.payload = }')

    async def confirm_responses(self):
        ''' '''
        start_time = time.time()
        base = 0
        completed = 0
        ready = 0
        shutdown = 0
        error = 0
        waiting = len(self.agent_collection)

        while waiting > 0:
            completed = 0
            ready = 0
            shutdown = 0
            error = 0
            waiting = 0
            num_did_step = 0
            for agent_id, _ in self.agent_collection.items():
                response = self.agent_collection[agent_id]['step_response'][self.time]
                if (response['reaction'] == 'completed'):
                    completed += 1
                elif (response['reaction'] == 'ready'):
                    ready += 1
                elif (response['reaction'] == 'error'):
                    error += 1
                elif (response['reaction'] == 'shutdown'):
                    shutdown += 1
                elif (response['reaction'] == 'waiting'):
                    waiting += 1

                if response['did_step']:
                    num_did_step += 1

            self.agent_stat[self.time] = {
                'completed': completed,
                'ready': ready,
                'error': error,
                'shutdown': shutdown,
                'waiting': waiting,
                'stepping_agents': num_did_step,
                'total_agents': len(self.agent_collection),
                # 'run_time_dist': []
            }
            current_time = time.time()
            if current_time - start_time >= 5:
                # logging.info(f"Waiting for Agent Response... {completed=}, {error=}, {shutdown=}, {waiting=} of {len(self.agent_collection)}: {base + (current_time - start_time):0.0f} sec")
                logging.info(f"Waiting for Agent Response... {self.agent_stat[self.time]}: {base + (current_time - start_time):0.0f} sec")
                base = base + (current_time - start_time)
                start_time = current_time

            await asyncio.sleep(0.1)

    async def step(self):

        logging.info(f"{self.scheduler_id} Step: {self.time}")
        start_time = time.time()

        if self.time == self.orsim_settings['SIMULATION_LENGTH_IN_STEPS']-1:
            message = {'action': 'shutdown', 'time_step': self.time}
        else:
            message = {'action': 'step', 'time_step': self.time}

        for agent_id, _ in self.agent_collection.items():
            self.agent_collection[agent_id]['step_response'][self.time] = {
                'reaction': 'waiting',
                'did_step': False,
                'run_time': 0
            }

        self.agent_messenger.client.publish(f'{self.run_id}/{self.scheduler_id}/ORSimAgent', json.dumps(message))

        try:
            # start_time = time.time()
            await asyncio.wait_for(self.confirm_responses(), timeout=self.orsim_settings['STEP_TIMEOUT'])
            # end_time = time.time()
            logging.info(f'{self.agent_stat[self.time] = }')
            # logging.info(f'{self.scheduler_id} Runtime: {(time.time()-start_time):0.2f} sec')

        except asyncio.TimeoutError as e:
            logging.exception(f'Scheduler {self.scheduler_id} timeout beyond {self.orsim_settings["STEP_TIMEOUT"] = } while waiting for confirm_responses.')
            self.step_timeout_handler(e)

        # Handle shutdown agents once successfully exiting the loop
        agents_shutdown = []
        for agent_id, agent_item in self.agent_collection.items():
            if agent_item['step_response'][self.time]['reaction'] in ['shutdown', 'waiting']:
                agents_shutdown.append(agent_id)

        for agent_id in agents_shutdown:
            self.remove_agent(agent_id)


        self.time += 1

        sim_stat = {
            'status': 'success',
            'end_sim': False,
        }

        if self.time == self.orsim_settings['SIMULATION_LENGTH_IN_STEPS']-1:
            sim_stat['end_sim'] = True


        logging.info(f'{self.scheduler_id} Runtime: {(time.time()-start_time):0.2f} sec')
        return sim_stat


    def step_timeout_handler(self, e):
        ''' '''
        tolerance = self.orsim_settings['STEP_TIMEOUT_TOLERANCE'] # Max % or agents having network issues


        if (self.agent_stat[self.time]['waiting'] / len(self.agent_collection)) <= tolerance:
            logging.warning(f"agent_stat = {self.pp.pformat(self.agent_stat[self.time])}")
            logging.warning(f"Unable to receive response from {self.agent_stat[self.time]['waiting']} Agents at {self.time=}. % Error ({(self.agent_stat[self.time]['waiting'] / len(self.agent_collection)):0.3f}) is within {tolerance=}. Continue processing...")
            # logging.warning(f'{self.pp.pformat(self.agent_collection)}')
        else:
            logging.error(f"Too many missing messages. % Error ({self.agent_stat[self.time]['waiting'] *100 / len(self.agent_collection)}) exceeded {tolerance=}. Abort...")
            logging.error(f'{self.pp.pformat(self.agent_collection)}')
            raise e




class CeleryConfig:

    # List of modules to import when the Celery worker starts.
    imports = ('apps.tasks',)

    ## Broker settings.
    broker_url = 'amqp://'
    ## Disable result backent and also ignore results.
    task_ignore_result = True




from abc import ABC, abstractclassmethod, abstractmethod
import asyncio, json, logging, time, os, traceback
from collections import OrderedDict

from datetime import datetime
from dateutil.relativedelta import relativedelta
from orsim.messenger import Messenger

from orsim.core.orsim_env import ORSimEnv
from orsim.utils import time_to_str, str_to_time

from cerberus import Validator

class ORSimAgent(ABC):

    messenger = None
    agent_failed = False
    payload_cache = None
    step_log = {}

    # def __init__(self, unique_id, run_id, reference_time, init_time_step, scheduler_id, behavior, orsim_settings):
    def __init__(self, unique_id, run_id, reference_time, init_time_step, scheduler, behavior): #, orsim_settings):
        self.unique_id = unique_id
        self.run_id = run_id
        self.scheduler_id = scheduler['id']
        self.reference_time = datetime.strptime(reference_time, '%Y%m%d%H%M%S') # datetime
        self.current_time = self.reference_time
        self.next_event_time = self.reference_time # To be set by agent at every step_response
        # self.orsim_settings = orsim_settings
        self.orsim_settings = ORSimEnv.validate_orsim_settings(scheduler['orsim_settings'])


        # Ideally behavior should be read from a datafile/db or in case of simulation, generated by the Model and passed in as attribute
        self.prev_time_step = 0
        self.current_time_step = 0
        self.elapsed_duration_steps = 0
        self.active = False

        self._shutdown = False
        self.behavior = behavior

        self.agent_credentials = {
            'email': f"{self.run_id}_{self.scheduler_id}_{unique_id}",
            'password': "secret_password",
        }

        self.messenger = Messenger(ORSimEnv.messenger_settings, self.agent_credentials)

        self.start_time = time.time()
        self.message_processing_active = False

        self.message_handlers = {}
        self.bootstrap_step(init_time_step)

        self.register_message_handler(topic=f"{self.run_id}/{self.scheduler_id}/ORSimAgent",
                                 method=self.handle_orsim_agent_message)

    def register_message_handler(self, topic, method):
        self.message_handlers[topic] = method

    def is_active(self):
        return self.active

    def reset_step_log(self):
        self.step_log = {}

    def add_step_log(self, message):
        self.step_log[datetime.now().isoformat()] = message

    def take_first_step(self, dummy_payload):
        self.message_processing_active = True
        self.handle_orsim_agent_message(dummy_payload)

    def handle_orsim_agent_message(self, payload):
        # print('Inside handle_orsim_agent_message')
        self.add_step_log('In handle_orsim_agent_message')

        try:
            self.bootstrap_step(payload['time_step'])

            if payload.get('action') == 'init':
                ''' NOTE This is unused block of code at the moment'''
                # print(f"{self.unique_id} received {payload=}")
                did_step = self.process_payload(payload)
                self.next_event_time = self.estimate_next_event_time()

                response_payload = {
                    'agent_id': self.unique_id,
                    'time_step': self.current_time_step,
                    'action': 'ready', # 'completed',
                    'did_step': did_step,
                    'run_time': time.time() - self.start_time,
                }
            elif payload.get('action') == 'step':

                did_step = self.process_payload(payload)
                self.next_event_time = self.estimate_next_event_time()

                response_payload = {
                    'agent_id': self.unique_id,
                    'time_step': self.current_time_step,
                    'action': 'completed' if self._shutdown==False else 'shutdown',
                    'did_step': did_step,
                    'run_time': time.time() - self.start_time,
                }
            elif payload.get('action') == 'shutdown':
                ''' '''
                # self.shutdown()
                response_payload = {
                    'agent_id': self.unique_id,
                    'time_step': self.current_time_step,
                    'action': 'shutdown',
                    'did_step': True,
                    'run_time': time.time() - self.start_time,
                }
        except Exception as e:
            response_payload = {
                'agent_id': self.unique_id,
                'time_step': self.current_time_step,
                'action': 'error',
                'did_step': False,
                'run_time': time.time() - self.start_time,
                'details': traceback.format_exc(), # str(e)
            }
            # logging.exception(f"{self.unique_id} raised {str(e)}")

        self.messenger.client.publish(f'{self.run_id}/{self.scheduler_id}/ORSimScheduler', json.dumps(response_payload))

        if payload.get('action') == 'shutdown':
            self.shutdown()

        self.end_time = time.time()
        self.message_processing_active = False

    def on_receive_message(self, client, userdata, message):
        ''' '''
        payload = json.loads(message.payload.decode('utf-8'))
        # print("received message", message.payload.decode('utf-8'))
        self.start_time = time.time()
        self.message_processing_active = True
        self.payload_cache = payload
        self.reset_step_log()

        logging.debug(f"Agent {self.unique_id} received {payload.get('action')}")

        if self.message_handlers.get(message.topic) is not None:
            method = self.message_handlers[message.topic]
            method(payload)

        logging.debug(f"Runtime for {self.unique_id} at {self.current_time_step}: {self.end_time - self.start_time:0.2f} secs ")

    def start_listening(self):
        start_time_for_ready = time.time() # NOTE Local start time NOT a class variable

        if not self.agent_failed:
            # self.agent_messenger = Messenger(self.agent_credentials, f"{self.run_id}/{self.scheduler_id}/ORSimAgent", self.on_receive_message)
            self.messenger.client.subscribe([(topic, 0) for topic, _ in self.message_handlers.items()])
            self.messenger.client.on_message = self.on_receive_message

            # print('subscribed to ', self.message_handlers)

            # if settings['CONCURRENCY_STRATEGY'] == 'ASYNCIO':
            #     logging.debug(f'Agent {self.unique_id} is Listening for Messages')
            #     loop = asyncio.get_event_loop()
            #     try:
            #         loop.run_forever()
            #     except KeyboardInterrupt:
            #         pass
            #     finally:
            #         loop.close()
            # elif settings['CONCURRENCY_STRATEGY'] == 'EVENTLET':
            import eventlet

            def run_forever():
                while True:
                    # eventlet.sleep(0.1)
                    eventlet.sleep(5)
                    self.handle_heartbeat_failure()
                    # logging.info(f"{self.unique_id} Heartbeat")
                    if self._shutdown == True:
                        self.stop_listening()
                        break

            eventlet.spawn(run_forever)

            # Once agent is setup and listening, send the ready message
            response_payload = {
                'agent_id': self.unique_id,
                'time_step': -1,
                'action': 'ready',
                'run_time': time.time() - start_time_for_ready,
            }
        else:
            response_payload = {
                'agent_id': self.unique_id,
                'time_step': -1,
                'action': 'init_error',
                'run_time': time.time() - start_time_for_ready,
            }

        self.take_first_step({
            'action': 'init',
            'time_step': self.current_time_step
        })

        # self.messenger.client.publish(f'{self.run_id}/{self.scheduler_id}/ORSimScheduler', json.dumps(response_payload))

    def stop_listening(self):
        self.messenger.disconnect()

    def get_current_time_str(self):
        return time_to_str(self.current_time)

    # @classmethod
    # def time_to_str(cls, time_var):
    #     return datetime.strftime(time_var, "%a, %d %b %Y %H:%M:%S GMT")

    # @classmethod
    # def str_to_time(cls, time_str):
    #     return datetime.strptime(time_str, "%a, %d %b %Y %H:%M:%S GMT")

    def bootstrap_step(self, time_step):

        self.prev_time_step = self.current_time_step
        self.current_time_step = time_step
        self.elapsed_duration_steps = self.current_time_step - self.prev_time_step

        self.current_time = self.reference_time + relativedelta(seconds = time_step * self.orsim_settings['STEP_INTERVAL'])

    def shutdown(self):
        if not self._shutdown:
            logging.info(f'Shutting down {self.unique_id = }')
            # self.stop_listening()
            self.logout()
            self.active = False
            self._shutdown = True

    def handle_heartbeat_failure(self):
        if self.message_processing_active:
            now = time.time()
            threshold = self.orsim_settings['STEP_TIMEOUT']
            if (now - self.start_time) > threshold:
                logging.warning(f"Auto Shutdown Agent {self.unique_id}. Exceeded heartbeat threshold {threshold} sec while processing...")
                # logging.warning(f"{self.payload_cache = }")
                logging.warning(f"{json.dumps(self.step_log, indent=2)}")
                # self.stop_listening()
                self.active = False
                self._shutdown = True

    def get_transition_probability(self, condition, default):
        try:
            for rule in self.behavior.get('transition_prob'):
                if rule[0] == condition:
                    return rule[1]
        except: pass

        return default

    @abstractmethod
    def process_payload(self, payload):
        raise NotImplementedError

    @abstractmethod
    def estimate_next_event_time(self):
        raise NotImplementedError

    @abstractmethod
    def logout(self):
        ''' process any logout processes needed in the agent.
        '''
        raise NotImplementedError





import json

# from numpy import isin
# from orsim.config import settings
import requests
# import urllib3
# from urllib.parse import quote
import logging

import paho.mqtt.client as paho


class Messenger:

    def __init__(self, settings, credentials, channel_id=None, on_message=None, transport=None):
        ''' '''
        self.settings = settings
        self.credentials = credentials
        self.channel_id = channel_id

        if transport is None:
            self.client = paho.Client(credentials['email'],  clean_session=True)
            self.client.username_pw_set(username=self.credentials['email'], password=self.credentials['password'])
            # Messenger.register_user(self.credentials['email'], self.credentials['password'])
            self.register_user(self.credentials['email'], self.credentials['password'])

            self.client.connect(self.settings['MQTT_BROKER'])

        if on_message is not None:
            self.client.on_message = on_message


        # RabbitMQ PubSub queue is used for processing requests in sequence
        # This is a deliberate design choice to enable:
        #   - Inter-Agent communication as core part of system design

        # if channel_id is not None:
        if isinstance(channel_id, str):
            # self.client.loop_start()
            self.client.subscribe(channel_id, qos=0)
            logging.debug(f"Channel: {channel_id}")
        elif isinstance(channel_id, list):
            # self.client.loop_start()
            self.client.subscribe([(cid, 0) for cid in channel_id])
            logging.debug(f"Channel: {channel_id}")

        self.client.loop_start()



    def disconnect(self):
        ''' '''
        # try:
        #     self.client.unsubscribe(self.channel_id)
        # except Exception as e:
        #     logging.exception(str(e))

        if self.channel_id is not None:
            try:
                self.client.loop_stop(force=True)
            except Exception as e:
                logging.exception(str(e))

        try:
            self.client.disconnect()
        except Exception as e:
            logging.exception(str(e))




    # @classmethod
    def register_user(self, username, password):
        ''' '''

        response = requests.get(f"{self.settings['RABBITMQ_MANAGEMENT_SERVER']}/users/{username}")
        if (response.status_code >= 200) and (response.status_code <= 299):
            logging.warning('User is already registered')
        else:
            try:
                response = requests.put(f"{self.settings['RABBITMQ_MANAGEMENT_SERVER']}/users/{username}",
                                        data=json.dumps({'password': password, 'tags': ''}),
                                        headers={"content-type": "application/json"},
                                        auth=(self.settings['RABBITMQ_ADMIN_USER'], self.settings['RABBITMQ_ADMIN_PASSWORD'])
                                    )
            except Exception as e:
                logging.exception(str(e))
                raise e

        # reset the user and set appropriate permissions as needed
        quoted_slash = '%2F'
        response = requests.put(f"{self.settings['RABBITMQ_MANAGEMENT_SERVER']}/permissions/{quoted_slash}/{username}",
                        data=json.dumps({"username":username, "vhost":"/", "configure":".*", "write":".*", "read":".*"}),
                        headers={"content-type": "application/json"},
                        auth=(self.settings['RABBITMQ_ADMIN_USER'], self.settings['RABBITMQ_ADMIN_PASSWORD'])
                    )

        response = requests.put(f"{self.settings['RABBITMQ_MANAGEMENT_SERVER']}/topic-permissions/{quoted_slash}/{username}",
                        data=json.dumps({username: username, "vhost": "/", "exchange": "", "write": ".*", "read": ".*"}),
                        headers={"content-type": "application/json"},
                        auth=(self.settings['RABBITMQ_ADMIN_USER'], self.settings['RABBITMQ_ADMIN_PASSWORD'])
                    )




"""Top-level package for orsim."""

__author__ = """remacutetigisti"""
__email__ = 'rajiv@nus.edu.sg'
__version__ = '0.2.2'

from orsim.core import *
from orsim.messenger import *
from orsim.utils import *
